#undef __TARGET_PSA__
#undef __TARGET_V1__

#include <core.p4>
#include <v1model.p4>
#include "common/headers.p4"


struct metadata_t {
    bit<9>  ingress_port;
    bit<32> nexthop;
}

struct mac_learn_digest_data {
    bit<48> src_addr;
    bit<9>  ingress_port;
}

struct arp_digest_data {
    bit<32> ip;		// destination (or nexthop)
    bit<48> mac;	// own MAC address to be used as SHA in ARP request
}

// ---------------------------------------------------------------------------
// Ingress parser
// ---------------------------------------------------------------------------

parser MyIngressParser(packet_in pkt,
		       out header_t hdr,
		       inout metadata_t meta,
		       inout standard_metadata_t stm
		      ) {

    state start {
	transition parse_ethernet;
    }

    state parse_ethernet {
        pkt.extract(hdr.eth);
        transition select(hdr.eth.ether_type) {
            ETHERTYPE_IPV4: parse_ipv4;
            default: accept;  // TODO: add VLAN and MAC-in-MAC support later
        }
    }

    state parse_ipv4 {
        pkt.extract(hdr.ipv4);
        transition select(hdr.ipv4.frag_offset, 
                          hdr.ipv4.protocol, 
                          hdr.ipv4.ihl) {
            (0, IPPROTO_UDP, 5)  : parse_udp;
            default              : accept;
        }
    }
     
    state parse_udp {
        pkt.extract(hdr.udp);
        transition select(hdr.udp.dst_port) {
            UDP_PORT_GTPU  : parse_gtpu;
            default        : accept;
        }
    }

    state parse_gtpu {
        pkt.extract(hdr.gtp);
        pkt.extract(hdr.ipv4_u);		// FIXME:  now assuming IPv4 payload only
        transition accept;			// FIXME2: not parsing user's UDP/TCP, maybe add later
    }
}


// ---------------------------------------------------------------------------
// Ingress Pipeline
// ---------------------------------------------------------------------------

control MyIngress(inout header_t hdr,
		  inout metadata_t meta,
		  inout standard_metadata_t stm
		 ) {

/*    counter(64, CounterType.packets_and_bytes) rx;
    counter(64, CounterType.packets_and_bytes) tx;
    counter(64, CounterType.packets_and_bytes) mac_error;
    counter(64, CounterType.packets_and_bytes) mac_lookup_error;
    counter(64, CounterType.packets_and_bytes) arp_error;
    counter(64, CounterType.packets_and_bytes) ip_lookup_error;
    counter(64, CounterType.packets_and_bytes) gtp_teid_unknown;
    counter(64, CounterType.packets_and_bytes) gtp_teid_mismatch;
    counter(64, CounterType.packets_and_bytes) unsupported;*/

    // Stateless (local) variables
    ipv4_addr_t _nexthop4 = 0;    
    ipv4_addr_t _ue_ipv4_addr = 0;
    ipv4_addr_t _bst_ipv4_addr = 0;
    ipv4_addr_t _ip_dst = 0;
    bit<32> _teid = 0;
    mac_addr_t _my_mac = EPG_VIRT_MAC;

    // Other modules
    
    /**********************************************************************************/

    action nop() {}

    action drop() {
	mark_to_drop(stm);
        exit;
    }

    action send(bit<9> port){
        stm.egress_port = port;
    }

    /**********************************************************************************/

    /*** MAC LEARN ***/ // TODO: add t4p4s digest
    action mac_learn_notify() {
    }

    table mac_learn {
        key = {
            hdr.eth.src_addr : exact;
        }
        actions = { mac_learn_notify; nop; }
        size = 1024;
        default_action = mac_learn_notify;
    }

    /*** MAC FORWARD ***/
    action mac_bcast() {	// FIXME: TBD, currently only fix DMAC works
    }

    table mac_forward {
        key = {
            hdr.eth.dst_addr : exact;
        }
        actions = {
            send;
            @defaultonly mac_bcast;
        }
        const default_action = mac_bcast();
        size = 1024;
    }

    /**********************************************************************************/

    /*** IP FORWARDING ***/
    action set_mac(mac_addr_t dstMac, ether_type_t type) {
        hdr.eth.src_addr = EPG_VIRT_MAC;
	hdr.eth.dst_addr = dstMac;
        hdr.eth.ether_type = type;
    }

    action arp_miss_notify() {
        meta.nexthop = _nexthop4;
    }

    table arp {
        key = { _nexthop4 : exact; }
        actions = {
            set_mac;
            @defaultonly arp_miss_notify;
        }
        const default_action = arp_miss_notify();
        size = 64;
    }

    action route4(ipv4_addr_t nh) {
        _nexthop4 = nh;
    }

    table ipv4_forward {
        key = {
            _ip_dst : lpm;
        }
        actions = {
            route4;
        }
        size = 1024;
    }

    /**********************************************************************************/

    /*** GTP ***/
    action gtp_encapsulate(bit<32> t, ipv4_addr_t addr) {
        hdr.ipv4.total_len    = hdr.ipv4_u.total_len + 36;	// length including GTP+UDP header
        hdr.ipv4.protocol     = IPPROTO_UDP;			// GTP is UDP
	hdr.ipv4.src_addr     = EPG_IP;			        // our address
	hdr.ipv4.dst_addr     = addr;			        // base station address
        hdr.ipv4.hdr_checksum = 0;				// calculated later

        hdr.udp.setValid();
        hdr.udp.src_port      = 2152;
        hdr.udp.dst_port      = 2152;
        hdr.udp.length        = hdr.ipv4_u.total_len + 16;	// FIXME: doesn't take into account GTP options
        hdr.udp.checksum      = 0;

        hdr.gtp.setValid();
	hdr.gtp.teid           = t;
        hdr.gtp.version        = 0x01;
        hdr.gtp.pFlag          = 1;
        hdr.gtp.messageType    = 0xff;
	hdr.gtp.messageLength  = hdr.ipv4_u.total_len;
    }

    action gtp_decapsulate_normal() {
	hdr.ipv4.setInvalid();
        hdr.udp.setInvalid();
        hdr.gtp.setInvalid();
    }

    action set_tunnel_params(bit<32> t, bit<32> a) {
        _teid = t;
        _bst_ipv4_addr = a;
    }

    table ueip_to_teid {
        key = {
            _ue_ipv4_addr : exact;
        }
        actions = {
            set_tunnel_params;
        }
        size = 100000;
    }

    /**********************************************************************************/

    apply {
//	rx.count(0);
        if (!hdr.eth.isValid()) {
//            mac_error.count(0);
	    drop();
            exit;
	}
	else {
	    if (hdr.eth.dst_addr[47:32] != _my_mac[47:32]) {
//		mac_error.count(0);
		drop();
                exit;
	    }
            else if (hdr.eth.dst_addr[31:0] != _my_mac[31:0]) {
//		mac_error.count(0);
		drop();
                exit;
	    }
	}

	mac_learn.apply();

	if (hdr.gtp.isValid()) { // UL
	    _ue_ipv4_addr = hdr.ipv4_u.src_addr;
	}
	else { // DL
	    _ue_ipv4_addr = hdr.ipv4.dst_addr;
	    hdr.ipv4_u.setValid();
	    hdr.ipv4_u = hdr.ipv4;
	    hdr.udp_u.setValid();
            hdr.udp_u  = hdr.udp;
	}

	if (!ueip_to_teid.apply().hit) {
//	    gtp_teid_unknown.count(0);
	    drop();
	    exit;
	}

	if (hdr.gtp.isValid()) { // UL
	    if (_teid != hdr.gtp.teid) {
//		gtp_teid_mismatch.count(0);
		drop();
		exit;
	    }
	    gtp_decapsulate_normal();
	    _ip_dst = hdr.ipv4_u.dst_addr;
	}
	else { // DL
	    if (hdr.ipv4_u.isValid()) { // DL IPv4 UE payload
		gtp_encapsulate(_teid, _bst_ipv4_addr);
		_ip_dst = hdr.ipv4_x.dst_addr;
	    }
	}

	if (!ipv4_forward.apply().hit) {			// routing (sets _nexthop4)
//	    ip_lookup_error.count(0);
	    drop();
	}

	if (!arp.apply().hit) {					// ARP (sets DMAC)
//	    arp_error.count(0);
	}

	if (!mac_forward.apply().hit) {				// L2 FWD (sets egress port)
//	    mac_lookup_error.count(0);
	}
	else {
//	    tx.count(0);
	    exit;
	}
    }
}

// ---------------------------------------------------------------------------
// Ingress Deparser
// ---------------------------------------------------------------------------

control MyIngressDeparser(packet_out pkt,
			  in header_t hdr
			 ) {
    apply {
	pkt.emit(hdr);
    }
}

control MyVerifyChecksum(inout header_t hdr, inout metadata_t meta) {
    apply {
    }
}

control Ipv4ComputeChecksum(inout header_t hdr, inout metadata_t meta) {
    apply { /*
	   update_checksum(
	     hdr.ipv4.isValid(),
            { hdr.ipv4.version,
	          hdr.ipv4.ihl,
              hdr.ipv4.diffserv,
              hdr.ipv4.totalLen,
              hdr.ipv4.identification,
              hdr.ipv4.flags,
              hdr.ipv4.fragOffset,
              hdr.ipv4.ttl,
              hdr.ipv4.protocol,
              hdr.ipv4.srcAddr,
              hdr.ipv4.dstAddr },
            hdr.ipv4.hdrChecksum,
            HashAlgorithm.csum16);
   */
    }
}

control EmptyEgress(inout header_t hdr, inout metadata_t meta, inout standard_metadata_t stm) {
    apply {
    }
}

V1Switch(MyIngressParser(), MyVerifyChecksum(), MyIngress(), EmptyEgress(), Ipv4ComputeChecksum(), MyIngressDeparser()) main;

