#undef __TARGET_PSA__
#undef __TARGET_V1__

#include <core.p4>

#ifdef __TARGET_TOFINO__
 #if __TARGET_TOFINO__ == 2
  #include <t2na.p4>
 #else
  #include <tna.p4>
 #endif
#else // x86: might be PSA or v1 model, select here
 #define USE_PSA 1
 #ifdef USE_PSA
  #include <psa.p4>
  #define __TARGET_PSA__ 1
 #else
  #include <v1model.p4>
  #define __TARGET_V1__ 1
 #endif
#endif

#include "common/headers.p4"
#ifdef __TARGET_TOFINO__
#include "common/util.p4"
#endif


struct metadata_t {
    bit<9>  ingress_port;
    bit<32> nexthop;
}

struct mac_learn_digest_data {
    bit<48> src_addr;
    bit<9>  ingress_port;
}

struct arp_digest_data {
    bit<32> ip;		// destination (or nexthop)
    bit<48> mac;	// own MAC address to be used as SHA in ARP request
}

// ---------------------------------------------------------------------------
// Ingress parser
// ---------------------------------------------------------------------------

parser MyIngressParser(packet_in pkt,
		       out header_t hdr,
		       inout metadata_t meta,
#ifdef __TARGET_TOFINO__
		       out ingress_intrinsic_metadata_t ig_intr_md
#endif
#ifdef __TARGET_PSA__
		       in psa_ingress_parser_input_metadata_t istd,
		       in EMPTY resubmit_meta,
		       in EMPTY recirculate_meta
#endif
#ifdef __TARGET_V1__
		       inout standard_metadata_t standard_metadata
#endif
		      ) {

    state start {
#ifdef __TARGET_TOFINO__
	pkt.extract(ig_intr_md);
        pkt.advance(PORT_METADATA_SIZE);
#endif
	transition parse_ethernet;
    }

    state parse_ethernet {
        pkt.extract(hdr.eth);
        transition select(hdr.eth.ether_type) {
            ETHERTYPE_IPV4: parse_ipv4;
            ETHERTYPE_IPV6: parse_ipv4;
            ETHERTYPE_ARP: parse_arp;
            default: accept;  // TODO: add VLAN and MAC-in-MAC support later
        }
    }

    state parse_arp {
        pkt.extract(hdr.arp);
        transition select(hdr.arp.htype, hdr.arp.ptype) {
            (ARP_HTYPE_ETHERNET, ARP_PTYPE_IPV4) : parse_arp2;
            default : accept;
        }
    }

    state parse_arp2 {
        transition select(hdr.arp.hlen, hdr.arp.plen) {
            (ARP_HLEN_ETHERNET,  ARP_PLEN_IPV4) : parse_arp_ipv4;
            default : accept;
        }
    }

    state parse_arp_ipv4 {
        pkt.extract(hdr.arp_ipv4);
        transition accept;
    }

    state parse_ipv4 {
        pkt.extract(hdr.ipv4);
        transition select(hdr.ipv4.frag_offset, 
                          hdr.ipv4.protocol, 
                          hdr.ipv4.ihl) {
            (0, IPPROTO_UDP, 5)  : parse_udp;
            (0, IPPROTO_ICMP, 5) : parse_icmp;
            default              : accept;
        }
    }
     
    state parse_ipv6 {
        pkt.extract(hdr.ipv6);
        transition select(hdr.ipv6.next_hdr) {
            IPPROTO_UDP  : parse_udp;
            IPPROTO_ICMP : parse_icmp;
            default      : accept;
        }
    }
     
    state parse_udp {
        pkt.extract(hdr.udp);
        transition select(hdr.udp.dst_port) {
            UDP_PORT_VXLAN : parse_vxlan;
            UDP_PORT_GTPU  : parse_gtpu;
            default        : accept;
        }
    }

    state parse_icmp {
        pkt.extract(hdr.icmp);
        transition accept;
    }

    state parse_gtpu {
        pkt.extract(hdr.gtp);
        pkt.extract(hdr.ipv4_u);		// FIXME:  now assuming IPv4 payload only
        transition accept;			// FIXME2: not parsing user's UDP/TCP, maybe add later
    }

    state parse_vxlan {
        pkt.extract(hdr.vxlan);
        transition parse_ethernet_2;
    }

    state parse_ethernet_2 {
        pkt.extract(hdr.eth_x);
        transition select(hdr.eth_x.ether_type) {
            ETHERTYPE_IPV4: parse_ipv4_x;
            ETHERTYPE_IPV6: parse_ipv6_x;
            ETHERTYPE_ARP: parse_arp;
            default: accept;  // TODO: add VLAN and MAC-in-MAC support later
        }
    }

    state parse_ipv4_x {
        pkt.extract(hdr.ipv4_x);
        transition select(hdr.ipv4_x.frag_offset,
                          hdr.ipv4_x.protocol,
                          hdr.ipv4_x.ihl) {
            (0, IPPROTO_UDP, 5)  : parse_udp_x;
            (0, IPPROTO_ICMP, 5) : parse_icmp;
            default              : accept;
        }
    }
     
    state parse_ipv6_x {
        pkt.extract(hdr.ipv6_x);
        transition select(hdr.ipv6_x.next_hdr) {
            IPPROTO_UDP  : parse_udp_x;
            IPPROTO_ICMP : parse_icmp;
            default      : accept;
        }
    }
     
    state parse_udp_x {
        pkt.extract(hdr.udp_x);
        transition select(hdr.udp_x.dst_port) {
            UDP_PORT_GTPU  : parse_gtpu;
            default        : accept;
        }
    }
}


// ---------------------------------------------------------------------------
// Ingress Pipeline
// ---------------------------------------------------------------------------

control MyIngress(inout header_t hdr,
		  inout metadata_t meta,
#ifdef __TARGET_TOFINO__
		  in ingress_intrinsic_metadata_t ig_intr_md,
		  in ingress_intrinsic_metadata_from_parser_t ig_prsr_md,
		  inout ingress_intrinsic_metadata_for_deparser_t ig_dprsr_md,
		  inout ingress_intrinsic_metadata_for_tm_t ig_tm_md
#endif
#ifdef __TARGET_PSA__
		  in psa_ingress_input_metadata_t istd,
		  inout psa_ingress_output_metadata_t ostd
#endif
#ifdef __TARGET_V1__
		  inout standard_metadata_t stm
#endif
		 ) {

#ifndef __TARGET_V1__
    // Stateful structures
    Counter<bit<64>, bit<1>>(1, CounterType_t.PACKETS_AND_BYTES) rx;	// TODO: do a per port ctr later
    Counter<bit<64>, bit<1>>(1, CounterType_t.PACKETS_AND_BYTES) tx;
    Counter<bit<32>, bit<1>>(1, CounterType_t.PACKETS) mac_error;
    Counter<bit<32>, bit<1>>(1, CounterType_t.PACKETS) mac_lookup_error;
    Counter<bit<32>, bit<1>>(1, CounterType_t.PACKETS) arp_error;
    Counter<bit<32>, bit<1>>(1, CounterType_t.PACKETS) ip_lookup_error;
    Counter<bit<32>, bit<1>>(1, CounterType_t.PACKETS) gtp_teid_unknown;
    Counter<bit<32>, bit<1>>(1, CounterType_t.PACKETS) gtp_teid_mismatch;
    Counter<bit<32>, bit<1>>(1, CounterType_t.PACKETS) unsupported;
#else
    counter(64, CounterType.packets_and_bytes) rx;
    counter(64, CounterType.packets_and_bytes) tx;
    counter(64, CounterType.packets_and_bytes) mac_error;
    counter(64, CounterType.packets_and_bytes) mac_lookup_error;
    counter(64, CounterType.packets_and_bytes) arp_error;
    counter(64, CounterType.packets_and_bytes) ip_lookup_error;
    counter(64, CounterType.packets_and_bytes) gtp_teid_unknown;
    counter(64, CounterType.packets_and_bytes) gtp_teid_mismatch;
    counter(64, CounterType.packets_and_bytes) unsupported;
#endif

    // Stateless (local) variables
    ipv4_addr_t _nexthop4 = 0;    
    ipv4_addr_t _ue_ipv4_addr = 0;
    ipv4_addr_t _bst_ipv4_addr = 0;
    ipv4_addr_t _ip_dst = 0;
    bit<32> _teid = 0;
    mac_addr_t _my_mac = EPG_VIRT_MAC;

    // Other modules
    
    /**********************************************************************************/

    action nop() {}

    action drop() {
#ifdef __TARGET_TOFINO__
	ig_dprsr_md.drop_ctl = 1;
#endif
#ifdef __TARGET_PSA__
	ostd.drop = true;
#endif
#ifdef __TARGET_V1__
	mark_to_drop(stm);
#endif
        exit;
    }

    action send(PortId_t port){
#ifdef __TARGET_TOFINO__
	ig_tm_md.ucast_egress_port = port;
#endif
#ifdef __TARGET_PSA__
        ostd.drop = false;
	ostd.egress_port = port;
#endif
#ifdef __TARGET_V1__
        stm.egress_port = port;
#endif
    }

    /**********************************************************************************/

    /*** MAC LEARN ***/ // TODO: add t4p4s digest
    action mac_learn_notify() {
#ifdef __TARGET_TOFINO__
        hdr.digest_hack.mac_addr = hdr.eth.src_addr;
        meta.ingress_port = ig_intr_md.ingress_port;
        ig_dprsr_md.digest_type = MAC_LEARN_DIGEST;
#else
/*    @name(".mac_learn") action mac_learn() {
        digest<mac_learn_digest>(MAC_LEARN_RECEIVER, { hdr.ethernet.srcAddr, standard_metadata.ingress_port } );
	meta.routing_metadata.mac_da = hdr.inner_ethernet.dstAddr;
        meta.routing_metadata.mac_sa = hdr.inner_ethernet.srcAddr;
    }*/
#endif
    }

    table mac_learn {
        key = {
            hdr.eth.src_addr : exact;
        }
        actions = { mac_learn_notify; nop; }
        size = 1024;
        default_action = mac_learn_notify;
    }

    /*** MAC FORWARD ***/
    action mac_bcast() {	// FIXME: TBD, currently only fix DMAC works
    }

    table mac_forward {
        key = {
            hdr.eth.dst_addr : exact;
        }
        actions = {
            send;
            @defaultonly mac_bcast;
        }
        const default_action = mac_bcast();
        size = 1024;
    }

    /**********************************************************************************/

    @name(".arp_reply") action arp_reply() {
        hdr.eth.dst_addr = hdr.arp_ipv4.sha;
        hdr.eth.src_addr = EPG_VIRT_MAC;
        hdr.arp.oper     = ARP_OPER_REPLY;
        hdr.arp_ipv4.tha = hdr.arp_ipv4.sha;
        hdr.arp_ipv4.tpa = hdr.arp_ipv4.spa;
        hdr.arp_ipv4.sha = EPG_VIRT_MAC;
        hdr.arp_ipv4.spa = EPG_IP;
    }

    @name(".send_icmp_reply") action send_icmp_reply() {
        bit<48>   tmp_mac;
        bit<32>   tmp_ip;
        tmp_mac               = hdr.eth.dst_addr;
        hdr.eth.dst_addr      = hdr.eth.src_addr;
        hdr.eth.src_addr      = tmp_mac;
        tmp_ip                = hdr.ipv4.dst_addr;
        hdr.ipv4.dst_addr     = hdr.ipv4.src_addr;
        hdr.ipv4.src_addr     = tmp_ip;
        hdr.icmp.type         = ICMP_ECHO_REPLY;
        hdr.icmp.hdr_checksum = 0; // For now
    }

    /**********************************************************************************/

    /*** IP FORWARDING ***/
    action set_mac(mac_addr_t dstMac, ether_type_t type) {
        hdr.eth.src_addr = EPG_VIRT_MAC;
	hdr.eth.dst_addr = dstMac;
        hdr.eth.ether_type = type;
    }

    action arp_miss_notify() {
        meta.nexthop = _nexthop4;
#ifdef __TARGET_TOFINO__
        ig_dprsr_md.digest_type = ARP_MISS_DIGEST;
#else
/*    @name(".mac_learn") action mac_learn() {
        digest<mac_learn_digest>(MAC_LEARN_RECEIVER, { hdr.ethernet.srcAddr, standard_metadata.ingress_port } );
	meta.routing_metadata.mac_da = hdr.inner_ethernet.dstAddr;
        meta.routing_metadata.mac_sa = hdr.inner_ethernet.srcAddr;
    }*/
#endif
    }

    table arp {
        key = { _nexthop4 : exact; }
        actions = {
            set_mac;
            @defaultonly arp_miss_notify;
        }
        const default_action = arp_miss_notify();
        size = 64;
    }

    action route4(ipv4_addr_t nh) {
        _nexthop4 = nh;
    }

    table ipv4_forward {
        key = {
            _ip_dst : lpm;
        }
        actions = {
            route4;
        }
        size = 1024;
    }

    /**********************************************************************************/

    /*** GTP ***/
    action gtp_encapsulate(bit<32> t, ipv4_addr_t addr) {
	hdr.ipv4_x.setValid();
        hdr.ipv4_x              = hdr.ipv4_u;			// basic parameters are the same
        hdr.ipv4_x.total_len    = hdr.ipv4_u.total_len + 36;	// length including GTP+UDP header
        hdr.ipv4_x.protocol     = IPPROTO_UDP;			// GTP is UDP
        hdr.ipv4_x.src_addr     = EPG_IP;			// our address
	hdr.ipv4_x.dst_addr     = addr;			        // base station address
        hdr.ipv4_x.hdr_checksum = 0;				// calculated later

        hdr.udp_x.setValid();
        hdr.udp_x.src_port      = 2152;
        hdr.udp_x.dst_port      = 2152;
        hdr.udp_x.length        = hdr.ipv4_u.total_len + 16;	// FIXME: doesn't take into account GTP options
        hdr.udp_x.checksum      = 0;

        hdr.gtp.setValid();
	hdr.gtp.teid             = t;
        hdr.gtp.version          = 0x01;
        hdr.gtp.pFlag            = 1;
        hdr.gtp.messageType      = 0xff;
	hdr.gtp.messageLength    = hdr.ipv4_u.total_len;

        hdr.udp_u = hdr.udp;
        hdr.udp.setInvalid();
    }

    action gtp_decapsulate_normal() {
	hdr.ipv4.setInvalid();
        hdr.udp.setInvalid();
        hdr.gtp.setInvalid();
    }

    action gtp_decapsulate_vxlan() {
        hdr.ipv4_x.setInvalid();
        hdr.udp_x.setInvalid();
        hdr.gtp.setInvalid();
    }

    action vxlan4_set_reply() {  // changing src with dst, the rest is not touched
        bit<32> tmp;
	tmp = hdr.ipv4.src_addr;
	hdr.ipv4.src_addr = hdr.ipv4.dst_addr;
	hdr.ipv4.dst_addr = tmp;
    }

/*    action vxlan_decapsulate() {
        hdr.ipv4.setInvalid();
	hdr.udp.setInvalid();
        hdr.vxlan.setInvalid();
	hdr.eth_x.setInvalid();
    }

    action vxlan_encapsulate(bit<24> vid, ipv4_addr_t addr) {
	hdr.vxlan.setValid();
        hdr.vxlan.vni = vid;
	hdr.eth_x.setValid();
	hdr.eth_x = hdr.eth;
	hdr.ipv4_x = hdr.ipv4;
        hdr.udp_x = hdr.udp;
	hdr.ipv4.dst_addr = addr;
	hdr.ipv4.src_addr = EPG_IP;
	hdr.udp.setValid();
        hdr.udp.dst_port = UDP_PORT_VXLAN;
    }
*/
    action set_tunnel_params(bit<32> t, bit<32> a) {
        _teid = t;
        _bst_ipv4_addr = a;
    }

    table ueip_to_teid {
        key = {
            _ue_ipv4_addr : exact;
        }
        actions = {
            set_tunnel_params;
        }
        size = 100000;
    }

    /**********************************************************************************/

    apply {
	rx.count(0);
        if (!hdr.eth.isValid()) {
            mac_error.count(0);
	    drop();
            exit;
	}
	else {
	    if (hdr.eth.dst_addr[47:32] != _my_mac[47:32]) {
		mac_error.count(0);
		drop();
                exit;
	    }
            else if (hdr.eth.dst_addr[31:0] != _my_mac[31:0]) {
		mac_error.count(0);
		drop();
                exit;
	    }
	}

	mac_learn.apply();

	if (hdr.arp.isValid()) {				// handle ARP requests
	    if (hdr.arp.oper == ARP_OPER_REQUEST) {
		if (hdr.arp_ipv4.isValid()) {
		    if (hdr.arp_ipv4.tpa == EPG_IP) {
			arp_reply();
#ifdef __TARGET_TOFINO__
			send(ig_intr_md.ingress_port);
#endif
#ifdef __TARGET_PSA__
			send(istd.ingress_port);
#endif
#ifdef __TARGET_V1__
			send(stm.ingress_port);
#endif
			tx.count(0);
			exit;
		    }
		}
	    }
	    drop();						// discard other ARP packets
	}
	else {

	    if (hdr.vxlan.isValid()) {
		if (hdr.gtp.isValid()) { // UL & VxLAN
		    _ue_ipv4_addr = hdr.ipv4_u.src_addr;
		}
		else { // DL & VxLAN
		    _ue_ipv4_addr = hdr.ipv4_x.dst_addr;
		    hdr.ipv4_u.setValid();
		    hdr.ipv4_u = hdr.ipv4_x; // TODO? DL UDP packets??? Might need to copy UDP (hdr.udp_x) as well...
		}
	    }
	    else {
		if (hdr.gtp.isValid()) { // UL & no VxLAN
		    _ue_ipv4_addr = hdr.ipv4_u.src_addr;
		}
		else { // DL & no VxLAN
                    _ue_ipv4_addr = hdr.ipv4.dst_addr;
		    hdr.ipv4_u.setValid();
                    hdr.ipv4_u = hdr.ipv4;
		    hdr.ipv4.setInvalid(); // TODO? DL UDP packets??? Might need to copy UDP (hdr.udp) as well...
		}
	    }

	    if (!ueip_to_teid.apply().hit) {
		gtp_teid_unknown.count(0);
		drop();
                exit;
	    }

	    if (hdr.gtp.isValid()) { // UL
		if (_teid != hdr.gtp.teid) {
		    gtp_teid_mismatch.count(0);
		    drop();
                    exit;
		}
		if (hdr.vxlan.isValid()) {
		    gtp_decapsulate_vxlan();
		}
		else{
		    gtp_decapsulate_normal();
		    _ip_dst = hdr.ipv4_u.dst_addr;
		}
	    }
	    else {
		if (hdr.ipv4_u.isValid()) { // DL IPv4 UE payload
		    gtp_encapsulate(_teid, _bst_ipv4_addr);
		    _ip_dst = hdr.ipv4_x.dst_addr;
		}
		else if (hdr.ipv6.isValid()) {
		    unsupported.count(0);
		    drop();
                    exit;
		}
	    }

	    if (hdr.vxlan.isValid()) {
		if (hdr.ipv4.isValid()) {
		    vxlan4_set_reply();
		    _ip_dst = hdr.ipv4.dst_addr;
		}
	    }

	    if (!ipv4_forward.apply().hit) {			// routing (sets _nexthop4)
		ip_lookup_error.count(0);
		drop();
	    }

	    if (!arp.apply().hit) {					// ARP (sets DMAC)
		arp_error.count(0);
	    }

	    if (!mac_forward.apply().hit) {				// L2 FWD (sets egress port)
		mac_lookup_error.count(0);
	    }
	    else {
		tx.count(0);
		exit;
	    }
	}
    }
}

// ---------------------------------------------------------------------------
// Ingress Deparser
// ---------------------------------------------------------------------------

control MyIngressDeparser(packet_out pkt,
#ifdef __TARGET_TOFINO__
			  inout header_t hdr,
			  in metadata_t meta,
			  in ingress_intrinsic_metadata_for_deparser_t ig_dprsr_md
#endif
#ifdef __TARGET_PSA__
			  out EMPTY clone_i2e_meta,
			  out EMPTY resubmit_meta,
			  out EMPTY normal_meta,
			  inout header_t hdr,
			  in metadata_t meta,
			  in psa_ingress_output_metadata_t istd
#endif
#ifdef __TARGET_V1__
			  in header_t hdr
#endif
			 ) {
#ifdef __TARGET_TOFINO__
    Digest <mac_learn_digest_data>() mac_learn_digest;
    Digest <arp_digest_data>() arp_digest;
    Checksum() cs;
#endif

    apply {
#ifdef __TARGET_TOFINO__
	if (ig_dprsr_md.digest_type == MAC_LEARN_DIGEST) {
            mac_learn_digest.pack({hdr.digest_hack.mac_addr, meta.ingress_port});
        }

	if (ig_dprsr_md.digest_type == ARP_MISS_DIGEST) {
            arp_digest.pack({meta.nexthop, hdr.eth.dst_addr});
        }

	if (hdr.vxlan.isValid()) {
	    hdr.ipv4.hdr_checksum = cs.update({
		hdr.ipv4.version, hdr.ipv4.ihl, hdr.ipv4.diffserv,
		hdr.ipv4.total_len, hdr.ipv4.identification, hdr.ipv4.flags,
		hdr.ipv4.frag_offset, hdr.ipv4.ttl, hdr.ipv4.protocol,
		hdr.ipv4.src_addr, hdr.ipv4.dst_addr});
	}
	if (hdr.ipv4_x.isValid()) {
	    hdr.ipv4_x.hdr_checksum = cs.update({
		hdr.ipv4_x.version, hdr.ipv4_x.ihl, hdr.ipv4_x.diffserv,
		hdr.ipv4_x.total_len, hdr.ipv4_x.identification, hdr.ipv4_x.flags,
		hdr.ipv4_x.frag_offset, hdr.ipv4_x.ttl, hdr.ipv4_x.protocol,
		hdr.ipv4_x.src_addr, hdr.ipv4_x.dst_addr});
	}
#endif
#ifdef __TARGET_PSA__
/*	if (hdr.vxlan.isValid()) {
	    cs.clear();
	    cs.add({hdr.ipv4.version, hdr.ipv4.ihl, hdr.ipv4.diffserv,hdr.ipv4.total_len, hdr.ipv4.identification, hdr.ipv4.flags,
	            hdr.ipv4.frag_offset, hdr.ipv4.ttl, hdr.ipv4.protocol, hdr.ipv4.src_addr, hdr.ipv4.dst_addr});
	    hdr.ipv4.hdr_checksum = cs.get();
	}
	if (hdr.ipv4_x.isValid()) {
	    cs.clear();
	    cs.add({hdr.ipv4_x.version, hdr.ipv4_x.ihl, hdr.ipv4_x.diffserv,hdr.ipv4_x.total_len, hdr.ipv4_x.identification, hdr.ipv4_x.flags,
                    hdr.ipv4_x.frag_offset, hdr.ipv4_x.ttl, hdr.ipv4_x.protocol, hdr.ipv4_x.src_addr, hdr.ipv4_x.dst_addr});
	    hdr.ipv4_x.hdr_checksum = cs.get();
        }*/
#endif

	pkt.emit(hdr);
    }
}

#ifdef __TARGET_TOFINO__
parser MyEgressParser(
    packet_in pkt,
    out empty_header_t hdr,
    out empty_metadata_t eg_md,
    out egress_intrinsic_metadata_t eg_intr_md) {

    TofinoEgressParser() tofino_parser;

    state start {
        tofino_parser.apply(pkt, eg_intr_md);
        transition accept;
    }

}

Pipeline(MyIngressParser(),
         MyIngress(),
         MyIngressDeparser(),
         MyEgressParser(),
         EmptyEgress(),
	 EmptyEgressDeparser()
	) pipe;
Switch(pipe) main;
#endif // TOFINO

#ifdef __TARGET_PSA__
parser EmptyEgressParser(packet_in buffer, out EMPTY a, inout EMPTY b, in psa_egress_parser_input_metadata_t c, in EMPTY d, in EMPTY e, in EMPTY f) {
    state start {
        transition accept;
    }
}

control EmptyEgress(inout EMPTY a, inout EMPTY b, in psa_egress_input_metadata_t c, inout psa_egress_output_metadata_t d) {
    apply {
    }
}

control EmptyEgressDeparser(packet_out buffer, out EMPTY a, out EMPTY b, inout EMPTY c, in EMPTY d, in psa_egress_output_metadata_t e, in psa_egress_deparser_input_metadata_t f) {
    apply {
    }
}

IngressPipeline(MyIngressParser(),
		MyIngress(),
		MyIngressDeparser()) ipipe;
EgressPipeline(EmptyEgressParser(),
	       EmptyEgress(),
	       EmptyEgressDeparser()) epipe;
PSA_Switch(ipipe, PacketReplicationEngine(), epipe, BufferingQueueingEngine()) main;
#endif // PSA

#ifdef __TARGET_V1__
control MyVerifyChecksum(inout header_t hdr, inout metadata_t meta) {
    apply {
    }
}

control Ipv4ComputeChecksum(inout header_t hdr, inout metadata_t meta) {
    apply { /*
	   update_checksum(
	     hdr.ipv4.isValid(),
            { hdr.ipv4.version,
	          hdr.ipv4.ihl,
              hdr.ipv4.diffserv,
              hdr.ipv4.totalLen,
              hdr.ipv4.identification,
              hdr.ipv4.flags,
              hdr.ipv4.fragOffset,
              hdr.ipv4.ttl,
              hdr.ipv4.protocol,
              hdr.ipv4.srcAddr,
              hdr.ipv4.dstAddr },
            hdr.ipv4.hdrChecksum,
            HashAlgorithm.csum16);
   */
    }
}

control EmptyEgress(inout header_t hdr, inout metadata_t meta, inout standard_metadata_t stm) {
    apply {
    }
}

V1Switch(MyIngressParser(), MyVerifyChecksum(), MyIngress(), EmptyEgress(), Ipv4ComputeChecksum(), MyIngressDeparser()) main;
#endif // V1MODEL
